#!/usr/bin/env python
# cardinal_pythonlib/interval.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Time interval classes and related functions.**

"""

# =============================================================================
# Imports
# =============================================================================

import datetime
import logging
import sys
from typing import List, Optional, Set, Tuple, Union
import unittest

from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger

log = logging.getLogger(__name__)

# =============================================================================
# Constants
# =============================================================================

SECONDS_PER_MINUTE = 60
MINUTES_PER_HOUR = 60
HOURS_PER_DAY = 24
DAYS_PER_WEEK = 7
DAYS_PER_YEAR = 365  # approx...

SECONDS_PER_HOUR = SECONDS_PER_MINUTE * MINUTES_PER_HOUR
SECONDS_PER_DAY = SECONDS_PER_HOUR * HOURS_PER_DAY
SECONDS_PER_WEEK = SECONDS_PER_DAY * DAYS_PER_WEEK
SECONDS_PER_YEAR = SECONDS_PER_DAY * DAYS_PER_YEAR  # approx...

NORMAL_DAY_START_H = 7
NORMAL_DAY_END_H = 19

BANK_HOLIDAYS = [datetime.datetime.strptime(x, "%Y-%m-%d").date() for x in [
    # https://www.gov.uk/bank-holidays
    # All bank holiday dates vary, even the date-based ones; e.g. if Christmas
    # Day is a Sunday, then the Christmas Day substitute bank holiday is Tue 27
    # Dec, after the Boxing Day Monday bank holiday.

    # 2014
    "2014-01-01",  # New Year's Day
    "2014-04-18",  # Good Friday
    "2014-04-21",  # Easter Monday
    "2014-05-05",  # Early May Bank Holiday
    "2014-05-26",  # Spring Bank Holiday
    "2014-08-25",  # Summer Bank Holiday
    "2014-12-25",  # Christmas Day
    "2014-12-26",  # Boxing Day
    # 2015
    "2015-01-01",  # New Year's Day
    "2015-04-03",  # Good Friday
    "2015-04-06",  # Easter Monday
    "2015-05-04",  # Early May Bank Holiday
    "2015-05-25",  # Spring Bank Holiday
    "2015-08-31",  # Summer Bank Holiday
    "2015-12-25",  # Christmas Day
    "2015-12-28",  # Boxing Day (substitute)
    # 2016
    "2016-01-01",  # New Year's Day
    "2016-03-25",  # Good Friday
    "2016-03-28",  # Easter Monday
    "2016-05-02",  # Early May Bank Holiday
    "2016-05-30",  # Spring Bank Holiday
    "2016-08-29",  # Summer Bank Holiday
    "2016-12-26",  # Boxing Day
    "2016-12-27",  # Christmas Day (substitute)
    # 2017
    "2017-01-02",  # New Year's Day (substitute day)
    "2017-04-14",  # Good Friday
    "2017-04-17",  # Easter Monday
    "2017-05-01",  # Early May bank holiday
    "2017-05-29",  # Spring bank holiday
    "2017-08-28",  # Summer bank holiday
    "2017-12-25",  # Christmas Day
    "2017-12-26",  # Boxing Day
    # 2018
    "2018-01-01",  # New Year's Day
    "2018-03-30",  # Good Friday
    "2018-04-02",  # Easter Monday
    "2018-05-07",  # Early May bank holiday
    "2018-05-28",  # Spring bank holiday
    "2018-08-28",  # Summer bank holiday
    "2018-12-25",  # Christmas Day
    "2018-12-26",  # Boxing Day
    # 2019
    "2019-01-01",  # New Year's Day
    "2019-04-19",  # Good Friday
    "2019-04-22",  # Easter Monday
    "2019-05-06",  # Early May bank holiday
    "2019-05-27",  # Spring bank holiday
    "2019-08-26",  # Summer bank holiday
    "2019-12-25",  # Christmas Day
    "2019-12-26",  # Boxing Day
    # 2020
    "2020-01-01",  # New Year's Day
    "2020-04-10",  # Good Friday
    "2020-04-13",  # Easter Monday
    "2020-05-08",  # Early May bank holiday (VE Day)
    "2020-05-25",  # Spring bank holiday
    "2020-08-31",  # Summer bank holiday
    "2020-12-25",  # Christmas Day
    "2020-12-28",  # Boxing Day (substitute day)
    # 2021
    "2020-01-01",  # New Year's Day
    "2020-04-02",  # Good Friday
    "2020-04-05",  # Easter Monday
    "2020-05-03",  # Early May bank holiday
    "2020-05-31",  # Spring bank holiday
    "2020-08-30",  # Summer bank holiday
    "2020-12-27",  # Christmas Day (substitute day)
    "2020-12-28",  # Boxing Day (substitute day)

    # Don't forget to add more in years to come.
]]
FIRST_KNOWN_BANK_HOLIDAY = min(x for x in BANK_HOLIDAYS)
LAST_KNOWN_BANK_HOLIDAY = max(x for x in BANK_HOLIDAYS)


# =============================================================================
# Helper functions
# =============================================================================

def formatdt(date: datetime.date, include_time: bool = True) -> str:
    """
    Formats a ``datetime.date`` to ISO-8601 basic format, to minute accuracy
    with no timezone (or, if ``include_time`` is ``False``, omit the time).
    """
    if include_time:
        return date.strftime("%Y-%m-%dT%H:%M")
    else:
        return date.strftime("%Y-%m-%d")


def convert_duration(duration: datetime.timedelta,
                     units: str) -> Optional[float]:
    """
    Convert a ``datetime.timedelta`` object -- a duration -- into other
    units. Possible units:

        ``s``, ``sec``, ``seconds``
        ``m``, ``min``, ``minutes``
        ``h``, ``hr``, ``hours``
        ``d``, ``days``
        ``w``, ``weeks``
        ``y``, ``years``
    """
    if duration is None:
        return None
    s = duration.total_seconds()
    if units in ['s', 'sec', 'seconds']:
        return s
    if units in ['m', 'min', 'minutes']:
        return s / SECONDS_PER_MINUTE
    if units in ['h', 'hr', 'hours']:
        return s / SECONDS_PER_HOUR
    if units in ['d', 'days']:
        return s / SECONDS_PER_DAY
    if units in ['w', 'weeks']:
        return s / SECONDS_PER_WEEK
    if units in ['y', 'years']:
        return s / SECONDS_PER_YEAR
    raise ValueError(f"Unknown units: {units}")


def is_uk_bank_holiday(date: datetime.date) -> bool:
    """
    Is the specified date (a ``datetime.date`` object) a UK bank holiday?

    Uses the ``BANK_HOLIDAYS`` list.
    """
    if date < FIRST_KNOWN_BANK_HOLIDAY:
        log.warning(f"Date {date} is earlier than first known bank holiday of "
                    f"{FIRST_KNOWN_BANK_HOLIDAY}; cardinal_pythonlib.interval "
                    f"may need updating")
    elif date > LAST_KNOWN_BANK_HOLIDAY:
        log.warning(f"Date {date} is later than last known bank holiday of "
                    f"{LAST_KNOWN_BANK_HOLIDAY}; cardinal_pythonlib.interval "
                    f"may need updating")
    return date in BANK_HOLIDAYS


def is_weekend(date: datetime.date) -> bool:
    """
    Is the specified date (a ``datetime.date`` object) a weekend?
    """
    return date.weekday() in [5, 6]


def is_saturday(date: datetime.date) -> bool:
    """
    Is the specified date (a ``datetime.date`` object) a Saturday?
    """
    return date.weekday() == 5


def is_sunday(date: datetime.date) -> bool:
    """
    Is the specified date (a ``datetime.date`` object) a Sunday?
    """
    return date.weekday() == 6


def is_normal_uk_working_day(date: datetime.date) -> bool:
    """
    Is the specified date (a ``datetime.date`` object) a normal working day,
    i.e. not a weekend or a bank holiday?
    """
    return not(is_weekend(date) or is_uk_bank_holiday(date))


# =============================================================================
# Interval
# =============================================================================

class Interval(object):
    """
    Object representing a time interval, with start and end objects that are
    normally ``datetime.datetime`` objects (though with care, a subset of some
    methods are possible with ``datetime.date`` objects; caveat emptor, and
    some methods will crash).

    Does not handle open-ended intervals (−∞, +∞) or null intervals.

    There's probably an existing class for this...
    """

    def __init__(self, start: datetime.datetime,
                 end: datetime.datetime) -> None:
        """
        Creates the interval.
        """
        if start is None or end is None:
            raise TypeError("Invalid interval creation")
        if start > end:
            (start, end) = (end, start)
        self.start = start
        self.end = end

    def __repr__(self) -> str:
        """
        Returns the canonical string representation of the object.
        """
        return f"Interval(start={self.start!r}, end={self.end!r})"

    def __str__(self) -> str:
        """
        Returns a string representation of the object.
        """
        return f"{formatdt(self.start)} − {formatdt(self.end)}"

    def __add__(self, value: datetime.timedelta) -> "Interval":
        """
        Adds a constant (``datetime.timedelta`` object) to the interval's start
        and end. Returns the new :class:`Interval`.
        """
        return Interval(self.start + value, self.end + value)

    def __lt__(self, other: "Interval") -> bool:
        """
        Allows sorting (on start time).
        """
        return self.start < other.start

    def copy(self) -> "Interval":
        """
        Returns a copy of the interval.
        """
        return Interval(self.start, self.end)

    def overlaps(self, other: "Interval") -> bool:
        """
        Does this interval overlap the other?

        Overlap:

        .. code-block:: none

            S--------S     S---S            S---S
              O---O          O---O        O---O

        Simpler method of testing is for non-overlap!

        .. code-block:: none

            S---S              S---S
                O---O      O---O
        """
        return not(self.end <= other.start or self.start >= other.end)

    def contiguous(self, other: "Interval") -> bool:
        """
        Does this interval overlap or touch the other?
        """
        return not(self.end < other.start or self.start > other.end)

    def contains(self, time: datetime.datetime,
                 inclusive: bool = True) -> bool:
        """
        Does the interval contain a momentary time?

        Args:
            time: the ``datetime.datetime`` to check
            inclusive: use inclusive rather than exclusive range checks?
        """
        if inclusive:
            return self.start <= time <= self.end
        else:
            return self.start < time < self.end

    def within(self, other: "Interval", inclusive: bool = True) -> bool:
        """
        Is this interval contained within the other?

        Args:
            other: the :class:`Interval` to check
            inclusive: use inclusive rather than exclusive range checks?
        """
        if not other:
            return False
        if inclusive:
            return self.start >= other.start and self.end <= other.end
        else:
            return self.start > other.start and self.end < other.end

    def union(self, other: "Interval") -> "Interval":
        """
        Returns an interval spanning the extent of this and the ``other``.
        """
        return Interval(
            min(self.start, other.start),
            max(self.end, other.end)
        )

    def intersection(self, other: "Interval") -> Optional["Interval"]:
        """
        Returns an :class:`Interval` representing the intersection of this and
        the ``other``, or ``None`` if they don't overlap.
        """
        if not self.contiguous(other):
            return None
        return Interval(
            max(self.start, other.start),
            min(self.end, other.end)
        )

    def cut(self, times: Union[datetime.datetime,
                               List[datetime.datetime]]) -> List["Interval"]:
        """
        Returns a list of intervals produced by using times (a list of
        ``datetime.datetime`` objects, or a single such object) as a set of
        knives to slice this interval.
        """
        if not isinstance(times, list):
            # Single time
            time = times
            if not self.contains(time):
                return []
            return [
                Interval(self.start, time),
                Interval(time, self.end)
            ]
        else:
            # Multiple times
            times = [t for t in times if self.contains(t)]  # discard others
            times.sort()
            times = [self.start] + times + [self.end]
            intervals = []
            for i in range(len(times) - 1):
                intervals.append(Interval(times[i], times[i + 1]))
            return intervals

    def duration(self) -> datetime.timedelta:
        """
        Returns a datetime.timedelta object representing the duration of this
        interval.
        """
        return self.end - self.start

    def duration_in(self, units: str) -> float:
        """
        Returns the duration of this interval in the specified units, as
        per :func:`convert_duration`.
        """
        return convert_duration(self.duration(), units)

    @staticmethod
    def wholeday(date: datetime.date) -> "Interval":
        """
        Returns an :class:`Interval` covering the date given (midnight at the
        start of that day to midnight at the start of the next day).
        """
        start = datetime.datetime.combine(date, datetime.time())
        return Interval(
            start,
            start + datetime.timedelta(days=1)
        )

    @staticmethod
    def daytime(date: datetime.date,
                daybreak: datetime.time = datetime.time(NORMAL_DAY_START_H),
                nightfall: datetime.time = datetime.time(NORMAL_DAY_END_H)) \
            -> "Interval":
        """
        Returns an :class:`Interval` representing daytime on the date given.
        """
        return Interval(
            datetime.datetime.combine(date, daybreak),
            datetime.datetime.combine(date, nightfall),
        )

    @staticmethod
    def dayspan(startdate: datetime.date,
                enddate: datetime.date,
                include_end: bool = True) -> Optional["Interval"]:
        """
        Returns an :class:`Interval` representing the date range given, from
        midnight at the start of the first day to midnight at the end of the
        last (i.e. at the start of the next day after the last), or if
        include_end is False, 24h before that.

        If the parameters are invalid, returns ``None``.
        """
        if enddate < startdate:
            return None
        if enddate == startdate and include_end:
            return None
        start_dt = datetime.datetime.combine(startdate, datetime.time())
        end_dt = datetime.datetime.combine(enddate, datetime.time())
        if include_end:
            end_dt += datetime.timedelta(days=1)
        return Interval(start_dt, end_dt)

    def component_on_date(self, date: datetime.date) -> Optional["Interval"]:
        """
        Returns the part of this interval that falls on the date given, or
        ``None`` if the interval doesn't have any part during that date.
        """
        return self.intersection(Interval.wholeday(date))

    def day_night_duration(
            self,
            daybreak: datetime.time = datetime.time(NORMAL_DAY_START_H),
            nightfall: datetime.time = datetime.time(NORMAL_DAY_END_H)) \
            -> Tuple[datetime.timedelta, datetime.timedelta]:
        """
        Returns a ``(day, night)`` tuple of ``datetime.timedelta`` objects
        giving the duration of this interval that falls into day and night
        respectively.
        """
        daytotal = datetime.timedelta()
        nighttotal = datetime.timedelta()
        startdate = self.start.date()
        enddate = self.end.date()
        ndays = (enddate - startdate).days + 1
        for i in range(ndays):
            date = startdate + datetime.timedelta(days=i)
            component = self.component_on_date(date)
            # ... an interval on a single day
            day = Interval.daytime(date, daybreak, nightfall)
            daypart = component.intersection(day)
            if daypart is not None:
                daytotal += daypart.duration()
                nighttotal += component.duration() - daypart.duration()
            else:
                nighttotal += component.duration()
        return daytotal, nighttotal

    def duration_outside_uk_normal_working_hours(
            self,
            starttime: datetime.time = datetime.time(NORMAL_DAY_START_H),
            endtime: datetime.time = datetime.time(NORMAL_DAY_END_H),
            weekdays_only: bool = False,
            weekends_only: bool = False) -> datetime.timedelta:
        """
        Returns a duration (a ``datetime.timedelta`` object) representing the
        number of hours outside normal working hours.

        This is not simply a subset of :meth:`day_night_duration`, because
        weekends are treated differently (they are always out of hours).

        The options allow the calculation of components on weekdays or weekends
        only.
        """
        if weekdays_only and weekends_only:
            raise ValueError("Can't have weekdays_only and weekends_only")
        ooh = datetime.timedelta()  # ooh = out of (normal) hours
        startdate = self.start.date()
        enddate = self.end.date()
        ndays = (enddate - startdate).days + 1
        for i in range(ndays):
            date = startdate + datetime.timedelta(days=i)
            component = self.component_on_date(date)
            # ... an interval on a single day
            if not is_normal_uk_working_day(date):
                if weekdays_only:
                    continue
                ooh += component.duration()  # all is out-of-normal-hours
            else:
                if weekends_only:
                    continue
                normalday = Interval.daytime(date, starttime, endtime)
                normalpart = component.intersection(normalday)
                if normalpart is not None:
                    ooh += component.duration() - normalpart.duration()
                else:
                    ooh += component.duration()
        return ooh

    def n_weekends(self) -> int:
        """
        Returns the number of weekends that this interval covers. Includes
        partial weekends.
        """
        startdate = self.start.date()
        enddate = self.end.date()
        ndays = (enddate - startdate).days + 1
        in_weekend = False
        n_weekends = 0
        for i in range(ndays):
            date = startdate + datetime.timedelta(days=i)
            if not in_weekend and is_weekend(date):
                in_weekend = True
                n_weekends += 1
            elif in_weekend and not is_weekend(date):
                in_weekend = False
        return n_weekends

    def saturdays_of_weekends(self) -> Set[datetime.date]:
        """
        Returns the dates of all Saturdays that are part of weekends that this
        interval covers (each Saturday representing a unique identifier for
        that weekend). The Saturday itself isn't necessarily the part of the
        weekend that the interval covers!
        """
        startdate = self.start.date()
        enddate = self.end.date()
        ndays = (enddate - startdate).days + 1
        saturdays = set()
        for i in range(ndays):
            date = startdate + datetime.timedelta(days=i)
            if is_saturday(date):
                saturdays.add(date)
            elif is_sunday(date):
                saturdays.add(date - datetime.timedelta(days=1))
        return saturdays


# =============================================================================
# IntervalList
# =============================================================================

class IntervalList(object):
    """
    Object representing a list of Intervals.
    Maintains an internally sorted state (by interval start time).
    """

    _ONLY_FOR_NO_INTERVAL = (
        "Only implemented for IntervalList objects with no_overlap == True"
    )

    # -------------------------------------------------------------------------
    # Constructor, representations, copying
    # -------------------------------------------------------------------------

    def __init__(self,
                 intervals: List[Interval] = None,
                 no_overlap: bool = True,
                 no_contiguous: bool = True) -> None:
        """
        Creates the :class:`IntervalList`.

        Args:
            intervals: optional list of :class:`Interval` objects to
                incorporate into the :class:`IntervalList`
            no_overlap: merge intervals that overlap (now and on subsequent
                addition)?
            no_contiguous: if ``no_overlap`` is set, merge intervals that are
                contiguous too?
        """
        # DO NOT USE intervals=[] in the function signature; that's the route
        # to a mutable default and a huge amount of confusion as separate
        # objects appear non-independent.
        self.intervals = [] if intervals is None else list(intervals)
        self.no_overlap = no_overlap
        self.no_contiguous = no_contiguous
        for i in self.intervals:
            if not isinstance(i, Interval):
                raise TypeError(
                    f"IntervalList creation failed: contents are not all "
                    f"Interval: {self.intervals!r}")
        self._tidy()

    def __repr__(self) -> str:
        """
        Returns the canonical string representation of the object.
        """
        return (
            f"IntervalList(intervals={self.intervals!r}, "
            f"no_overlap={self.no_overlap}, "
            f"no_contiguous={self.no_contiguous})")

    def copy(self, no_overlap: bool = None,
             no_contiguous: bool = None) -> "IntervalList":
        """
        Makes and returns a copy of the :class:`IntervalList`. The
        ``no_overlap``/``no_contiguous`` parameters can be changed.

        Args:
            no_overlap: merge intervals that overlap (now and on subsequent
                addition)?
            no_contiguous: if ``no_overlap`` is set, merge intervals that are
                contiguous too?
        """
        if no_overlap is None:
            no_overlap = self.no_overlap
        if no_contiguous is None:
            no_contiguous = self.no_contiguous
        return IntervalList(self.intervals, no_overlap=no_overlap,
                            no_contiguous=no_contiguous)

    def list(self) -> List[Interval]:
        """
        Returns the contained list of :class:`Interval` objects.
        """
        return self.intervals

    # -------------------------------------------------------------------------
    # Add an interval
    # -------------------------------------------------------------------------

    def add(self, interval: Interval) -> None:
        """
        Adds an interval to the list. If ``self.no_overlap`` is True, as is the
        default, it will merge any overlapping intervals thus created.
        """
        if interval is None:
            return
        if not isinstance(interval, Interval):
            raise TypeError(
                "Attempt to insert non-Interval into IntervalList")
        self.intervals.append(interval)
        self._tidy()

    # -------------------------------------------------------------------------
    # Internal consolidation functions, and sorting
    # -------------------------------------------------------------------------

    def _tidy(self) -> None:
        """
        Removes overlaps, etc., and sorts.
        """
        if self.no_overlap:
            self.remove_overlap(self.no_contiguous)  # will sort
        else:
            self._sort()

    def _sort(self) -> None:
        """
        Sorts (in place) by interval start time.
        """
        self.intervals.sort()

    def _remove_overlap_sub(self, also_remove_contiguous: bool) -> bool:
        """
        Called by :meth:`remove_overlap`. Removes the first overlap found.

        Args:
            also_remove_contiguous: treat contiguous (as well as overlapping)
                intervals as worthy of merging?

        Returns:
            bool: ``True`` if an overlap was removed; ``False`` otherwise

        """
        # Returns
        for i in range(len(self.intervals)):
            for j in range(i + 1, len(self.intervals)):
                first = self.intervals[i]
                second = self.intervals[j]
                if also_remove_contiguous:
                    test = first.contiguous(second)
                else:
                    test = first.overlaps(second)
                if test:
                    newint = first.union(second)
                    self.intervals.pop(j)
                    self.intervals.pop(i)  # note that i must be less than j
                    self.intervals.append(newint)
                    return True
        return False

    def remove_overlap(self, also_remove_contiguous: bool = False) -> None:
        """
        Merges any overlapping intervals.

        Args:
            also_remove_contiguous: treat contiguous (as well as overlapping)
                intervals as worthy of merging?
        """
        overlap = True
        while overlap:
            overlap = self._remove_overlap_sub(also_remove_contiguous)
        self._sort()

    def _any_overlap_or_contiguous(self, test_overlap: bool) -> bool:
        """
        Do any of the intervals overlap?

        Args:
            test_overlap: if ``True``, test for overlapping intervals; if
                ``False``, test for contiguous intervals.
        """
        for i in range(len(self.intervals)):
            for j in range(i + 1, len(self.intervals)):
                first = self.intervals[i]
                second = self.intervals[j]
                if test_overlap:
                    test = first.overlaps(second)
                else:
                    test = first.contiguous(second)
                if test:
                    return True
        return False

    # -------------------------------------------------------------------------
    # Simple descriptions
    # -------------------------------------------------------------------------

    def is_empty(self) -> bool:
        """
        Do we have no intervals?
        """
        return len(self.intervals) == 0

    def any_overlap(self) -> bool:
        """
        Do any of the intervals overlap?
        """
        return self._any_overlap_or_contiguous(test_overlap=True)

    def any_contiguous(self) -> bool:
        """
        Are any of the intervals contiguous?
        """
        return self._any_overlap_or_contiguous(test_overlap=False)

    # -------------------------------------------------------------------------
    # Start, end, range, duration
    # -------------------------------------------------------------------------

    def start_datetime(self) -> Optional[datetime.datetime]:
        """
        Returns the start date of the set of intervals, or ``None`` if empty.
        """
        if not self.intervals:
            return None
        return self.intervals[0].start
        # Internally sorted by start date, so this is always OK.

    def end_datetime(self) -> Optional[datetime.datetime]:
        """
        Returns the end date of the set of intervals, or ``None`` if empty.
        """
        if not self.intervals:
            return None
        return max([x.end for x in self.intervals])

    def start_date(self) -> Optional[datetime.date]:
        """
        Returns the start date of the set of intervals, or ``None`` if empty.
        """
        if not self.intervals:
            return None
        return self.start_datetime().date()

    def end_date(self) -> Optional[datetime.date]:
        """
        Returns the end date of the set of intervals, or ``None`` if empty.
        """
        if not self.intervals:
            return None
        return self.end_datetime().date()

    def extent(self) -> Optional[Interval]:
        """
        Returns an :class:`Interval` running from the earliest start of an
        interval in this list to the latest end. Returns ``None`` if we are
        empty.
        """
        if not self.intervals:
            return None
        return Interval(self.start_datetime(), self.end_datetime())

    def total_duration(self) -> datetime.timedelta:
        """
        Returns a ``datetime.timedelta`` object with the total sum of
        durations. If there is overlap, time will be double-counted, so beware!
        """
        total = datetime.timedelta()
        for interval in self.intervals:
            total += interval.duration()
        return total

    # -------------------------------------------------------------------------
    # Intervals and durations within our list
    # -------------------------------------------------------------------------

    def get_overlaps(self) -> "IntervalList":
        """
        Returns an :class:`IntervalList` containing intervals representing
        periods of overlap between intervals in this one.
        """
        overlaps = IntervalList()
        for i in range(len(self.intervals)):
            for j in range(i + 1, len(self.intervals)):
                first = self.intervals[i]
                second = self.intervals[j]
                ol = first.intersection(second)
                if ol is not None:
                    overlaps.add(ol)
        return overlaps

    def durations(self) -> List[datetime.timedelta]:
        """
        Returns a list of ``datetime.timedelta`` objects representing the
        durations of each interval in our list.
        """
        return [x.duration() for x in self.intervals]

    def longest_duration(self) -> Optional[datetime.timedelta]:
        """
        Returns the duration of the longest interval, or None if none.
        """
        if not self.intervals:
            return None
        return max(self.durations())

    def longest_interval(self) -> Optional[Interval]:
        """
        Returns the longest interval, or ``None`` if none.
        """
        longest_duration = self.longest_duration()
        for i in self.intervals:
            if i.duration() == longest_duration:
                return i
        return None

    def shortest_duration(self) -> Optional[datetime.timedelta]:
        """
        Returns the duration of the longest interval, or ``None`` if none.
        """
        if not self.intervals:
            return None
        return min(self.durations())

    def shortest_interval(self) -> Optional[Interval]:
        """
        Returns the shortest interval, or ``None`` if none.
        """
        shortest_duration = self.shortest_duration()
        for i in self.intervals:
            if i.duration() == shortest_duration:
                return i
        return None

    def first_interval_starting(self, start: datetime.datetime) -> \
            Optional[Interval]:
        """
        Returns our first interval that starts with the ``start`` parameter, or
        ``None``.
        """
        for i in self.intervals:
            if i.start == start:
                return i
        return None

    def first_interval_ending(self, end: datetime.datetime) \
            -> Optional[Interval]:
        """
        Returns our first interval that ends with the ``end`` parameter, or
        ``None``.
        """
        for i in self.intervals:
            if i.end == end:
                return i
        return None

    # -------------------------------------------------------------------------
    # Gaps and subsets
    # -------------------------------------------------------------------------

    def gaps(self) -> "IntervalList":
        """
        Returns all the gaps between intervals, as an :class:`IntervalList`.
        """
        if len(self.intervals) < 2:
            return IntervalList(None)
        gaps = []
        for i in range(len(self.intervals) - 1):
            gap = Interval(
                self.intervals[i].end,
                self.intervals[i + 1].start
            )
            gaps.append(gap)
        return IntervalList(gaps)

    def shortest_gap(self) -> Optional[Interval]:
        """
        Find the shortest gap between intervals, or ``None`` if none.
        """
        gaps = self.gaps()
        return gaps.shortest_interval()

    def shortest_gap_duration(self) -> Optional[datetime.timedelta]:
        """
        Find the duration of the shortest gap between intervals, or ``None`` if
        none.
        """
        gaps = self.gaps()
        return gaps.shortest_duration()

    def subset(self, interval: Interval,
               flexibility: int = 2) -> "IntervalList":
        """
        Returns an IntervalList that's a subset of this one, only containing
        intervals that meet the "interval" parameter criterion. What "meet"
        means is defined by the ``flexibility`` parameter.

        ``flexibility == 0``: permits only wholly contained intervals:

        .. code-block:: none

            interval:
                        I----------------I

            intervals in self that will/won't be returned:

                N---N  N---N   Y---Y   N---N   N---N
                    N---N                N---N

        ``flexibility == 1``: permits overlapping intervals as well:

        .. code-block:: none

                        I----------------I

                N---N  Y---Y   Y---Y   Y---Y   N---N
                    N---N                N---N

        ``flexibility == 2``: permits adjoining intervals as well:

        .. code-block:: none

                        I----------------I

                N---N  Y---Y   Y---Y   Y---Y   N---N
                    Y---Y                Y---Y
        """
        if flexibility not in [0, 1, 2]:
            raise ValueError("subset: bad flexibility value")
        permitted = []
        for i in self.intervals:
            if flexibility == 0:
                ok = i.start > interval.start and i.end < interval.end
            elif flexibility == 1:
                ok = i.end > interval.start and i.start < interval.end
            else:
                ok = i.end >= interval.start and i.start <= interval.end
            if ok:
                permitted.append(i)
        return IntervalList(permitted)

    def gap_subset(self, interval: Interval,
                   flexibility: int = 2) -> "IntervalList":
        """
        Returns an IntervalList that's a subset of this one, only containing
        *gaps* between intervals that meet the interval criterion.

        See :meth:`subset` for the meaning of parameters.
        """
        return self.gaps().subset(interval, flexibility)

    # -------------------------------------------------------------------------
    # Descriptions relating to the working week (for rota work)
    # -------------------------------------------------------------------------

    def n_weekends(self) -> int:
        """
        Returns the number of weekends that the intervals collectively touch
        (where "touching a weekend" means "including time on a Saturday or a
        Sunday").
        """
        saturdays = set()
        for interval in self.intervals:
            saturdays.update(interval.saturdays_of_weekends())
        return len(saturdays)

    def duration_outside_nwh(
            self,
            starttime: datetime.time = datetime.time(NORMAL_DAY_START_H),
            endtime: datetime.time = datetime.time(NORMAL_DAY_END_H)) \
            -> datetime.timedelta:
        """
        Returns the total duration outside normal working hours, i.e.
        evenings/nights, weekends (and Bank Holidays).
        """
        total = datetime.timedelta()
        for interval in self.intervals:
            total += interval.duration_outside_uk_normal_working_hours(
                starttime, endtime)
        return total

    def max_consecutive_days(self) -> Optional[Tuple[int, Interval]]:
        """
        The length of the longest sequence of days in which all days include
        an interval.

        Returns:
             tuple:
                ``(longest_length, longest_interval)`` where
                ``longest_interval`` is a :class:`Interval` containing the
                start and end date of the longest span -- or ``None`` if we
                contain no intervals.
        """
        if len(self.intervals) == 0:
            return None
        startdate = self.start_date()
        enddate = self.end_date()
        seq = ''
        ndays = (enddate - startdate).days + 1
        for i in range(ndays):
            date = startdate + datetime.timedelta(days=i)
            wholeday = Interval.wholeday(date)
            if any([x.overlaps(wholeday) for x in self.intervals]):
                seq += '+'
            else:
                seq += ' '
        # noinspection PyTypeChecker
        longest = max(seq.split(), key=len)
        longest_len = len(longest)
        longest_idx = seq.index(longest)
        longest_interval = Interval.dayspan(
            startdate + datetime.timedelta(days=longest_idx),
            startdate + datetime.timedelta(days=longest_idx + longest_len)
        )
        return longest_len, longest_interval

    def _sufficient_gaps(self,
                         startdate: datetime.date,
                         enddate: datetime.date,
                         requiredgaps: List[datetime.timedelta],
                         flexibility: int) -> Tuple[bool, Optional[Interval]]:
        """
        Are there sufficient gaps (specified by ``requiredgaps``) in the date
        range specified? This is a worker function for :meth:`sufficient_gaps`.
        """
        requiredgaps = list(requiredgaps)  # make a copy
        interval = Interval.dayspan(startdate, enddate, include_end=True)
        # log.debug(">>> _sufficient_gaps")
        gaps = self.gap_subset(interval, flexibility)
        gapdurations = gaps.durations()
        gaplist = gaps.list()
        gapdurations.sort(reverse=True)  # longest gap first
        requiredgaps.sort(reverse=True)  # longest gap first
        # log.debug("... gaps = {}".format(gaps))
        # log.debug("... gapdurations = {}".format(gapdurations))
        # log.debug("... requiredgaps = {}".format(requiredgaps))
        while requiredgaps:
            # log.debug("... processing gap")
            if not gapdurations:
                # log.debug("<<< no gaps left")
                return False, None
            if gapdurations[0] < requiredgaps[0]:
                # log.debug("<<< longest gap is too short")
                return False, self.first_interval_ending(gaplist[0].start)
            gapdurations.pop(0)
            requiredgaps.pop(0)
            gaplist.pop(0)
            # ... keeps gaplist and gapdurations mapped to each other
        # log.debug("<<< success")
        return True, None

    def sufficient_gaps(self,
                        every_n_days: int,
                        requiredgaps: List[datetime.timedelta],
                        flexibility: int = 2) \
            -> Tuple[bool, Optional[Interval]]:
        """
        Are gaps present sufficiently often?
        For example:

        .. code-block:: python

            every_n_days=21
            requiredgaps=[
                datetime.timedelta(hours=62),
                datetime.timedelta(hours=48),
            ]

        ... means "is there at least one 62-hour gap and one (separate) 48-hour
        gap in every possible 21-day sequence within the IntervalList?

        - If ``flexibility == 0``: gaps must be WHOLLY WITHIN the interval.

        - If ``flexibility == 1``: gaps may OVERLAP the edges of the interval.

        - If ``flexibility == 2``: gaps may ABUT the edges of the interval.

        Returns ``(True, None)`` or ``(False, first_failure_interval)``.
        """
        if len(self.intervals) < 2:
            return False, None
        startdate = self.start_date()
        enddate = self.end_date()
        ndays = (enddate - startdate).days + 1
        if ndays <= every_n_days:
            # Our interval is too short, or just right
            return self._sufficient_gaps(startdate, enddate, requiredgaps,
                                         flexibility)
        for i in range(ndays - every_n_days):
            j = i + every_n_days
            a = startdate + datetime.timedelta(days=i)
            b = startdate + datetime.timedelta(days=j)
            sufficient, ffi = self._sufficient_gaps(a, b, requiredgaps,
                                                    flexibility)
            if not sufficient:
                return False, ffi
        return True, None

    # -------------------------------------------------------------------------
    # Cumulative time calculations
    # -------------------------------------------------------------------------

    def cumulative_time_to(self,
                           when: datetime.datetime) -> datetime.timedelta:
        """
        Returns the cumulative time contained in our intervals up to the
        specified time point.
        """
        assert self.no_overlap, self._ONLY_FOR_NO_INTERVAL
        cumulative = datetime.timedelta()
        for interval in self.intervals:
            if interval.start >= when:
                break
            elif interval.end <= when:
                # complete interval precedes "when"
                cumulative += interval.duration()
            else:  # start < when < end
                cumulative += when - interval.start
        return cumulative

    def cumulative_gaps_to(self,
                           when: datetime.datetime) -> datetime.timedelta:
        """
        Return the cumulative time within our gaps, up to ``when``.
        """
        gaps = self.gaps()
        return gaps.cumulative_time_to(when)

    def time_afterwards_preceding(
            self, when: datetime.datetime) -> Optional[datetime.timedelta]:
        """
        Returns the time after our last interval, but before ``when``.
        If ``self`` is an empty list, returns ``None``.
        """
        if self.is_empty():
            return None
        end_time = self.end_datetime()
        if when <= end_time:
            return datetime.timedelta()
        else:
            return when - end_time

    def cumulative_before_during_after(self,
                                       start: datetime.datetime,
                                       when: datetime.datetime) -> \
            Tuple[datetime.timedelta,
                  datetime.timedelta,
                  datetime.timedelta]:
        """
        For a given time, ``when``, returns the cumulative time

        - after ``start`` but before ``self`` begins, prior to ``when``;
        - after ``start`` and during intervals represented by ``self``, prior
          to ``when``;
        - after ``start`` and after at least one interval represented by
          ``self`` has finished, and not within any intervals represented by
          ``self``, and prior to ``when``.

        Args:
            start: the start time of interest (e.g. before ``self`` begins)
            when: the time of interest
            
        Returns:
            tuple: ``before, during, after``

        Illustration
        
        .. code-block:: none
        
        
            start:      S
            self:           X---X       X---X       X---X       X---X
            
            when:                                           W
            
            before:     ----
            during:         -----       -----       -----
            after:               -------     -------     ----

        """
        assert self.no_overlap, (
            "Only implemented for IntervalList objects with no_overlap == True"
        )
        no_time = datetime.timedelta()
        earliest_interval_start = self.start_datetime()

        # Easy special cases
        if when <= start:
            return no_time, no_time, no_time
        if self.is_empty() or when <= earliest_interval_start:
            return when - start, no_time, no_time

        # Now we can guarantee:
        # - "self" is a non-empty list
        # - start < when
        # - earliest_interval_start < when

        # Before
        if earliest_interval_start < start:
            before = no_time
        else:
            before = earliest_interval_start - start

        # During
        during = self.cumulative_time_to(when)

        after = (
            self.cumulative_gaps_to(when) +
            self.time_afterwards_preceding(when)
        )

        return before, during, after


# =============================================================================
# Unit testing
# =============================================================================

class TestInterval(unittest.TestCase):
    """
    Unit tests.
    """
    def test_interval(self) -> None:
        a = datetime.datetime(2015, 1, 1)
        log.debug(f"a = {a!r}")
        b = datetime.datetime(2015, 1, 6)
        log.debug(f"b = {b!r}")
        i = Interval(a, b)
        log.debug(f"i = {i!r}")
        j = i + datetime.timedelta(hours=3)
        log.debug(f"j = {j!r}")
        cut = i.cut(datetime.datetime(2015, 1, 3))
        log.debug(f"cut = {cut!r}")


# =============================================================================
# main
# =============================================================================

if __name__ == "__main__":
    main_only_quicksetup_rootlogger(level=logging.DEBUG)
    log.info("Running unit tests")
    unittest.main(argv=[sys.argv[0]])
    sys.exit(0)



#!/usr/bin/env python
# cardinal_pythonlib/dogpile_cache.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Extensions to dogpile.cache.**

1.  The basic cache objects.

2.  FIX FOR DOGPILE.CACHE FOR DECORATED FUNCTIONS, 2017-07-28 (PLUS SOME OTHER
    IMPROVEMENTS). SEE

        https://bitbucket.org/zzzeek/dogpile.cache/issues/96/error-in-python-35-with-use-of-deprecated

    This fixes a crash using type-hinted functions under Python 3.5 with
    ``dogpile.cache==0.6.4``:

    .. code-block:: none

        Traceback (most recent call last):
          File "/usr/lib/python3.5/runpy.py", line 184, in _run_module_as_main
            "__main__", mod_spec)
          File "/usr/lib/python3.5/runpy.py", line 85, in _run_code
            exec(code, run_globals)
          File "/home/rudolf/Documents/code/camcops/server/camcops_server/cc_modules/cc_cache.py", line 64, in <module>
            unit_test_cache()
          File "/home/rudolf/Documents/code/camcops/server/camcops_server/cc_modules/cc_cache.py", line 50, in unit_test_cache
            def testfunc() -> str:
          File "/home/rudolf/dev/venvs/camcops/lib/python3.5/site-packages/dogpile/cache/region.py", line 1215, in decorator
            key_generator = function_key_generator(namespace, fn)
          File "/home/rudolf/dev/venvs/camcops/lib/python3.5/site-packages/dogpile/cache/util.py", line 31, in function_key_generator
            args = inspect.getargspec(fn)
          File "/usr/lib/python3.5/inspect.py", line 1045, in getargspec
            raise ValueError("Function has keyword-only arguments or annotations"
        ValueError: Function has keyword-only arguments or annotations, use getfullargspec() API which can support them

3.  Other improvements include:

    - the cache decorators operate as:
        - PER-INSTANCE caches for class instances, provided the first parameter
          is named "self";
        - PER-CLASS caches for classmethods, provided the first parameter is
          named "cls";
        - PER-FUNCTION caches for staticmethods and plain functions

    - keyword arguments are supported

    - properties are supported (the @property decorator must be ABOVE the
      cache decorator)

    - Note that this sort of cache relies on the generation of a STRING KEY
      from the function arguments. It uses the ``hex(id())`` function for
      ``self``/``cls`` arguments, and the ``to_str()`` function, passed as a
      parameter, for others (for which the default is ``"repr"``; see
      discussion below as to why ``"repr"`` is suitable while ``"str"`` is
      not).

"""  # noqa


# =============================================================================
# Imports; logging
# =============================================================================

import inspect
import logging
from typing import Any, Callable, Dict, List, Optional

# noinspection PyUnresolvedReferences,PyPackageRequirements
from dogpile.cache import make_region
# from dogpile.util import compat  # repr used as the default instead of compat.to_str  # noqa

from cardinal_pythonlib.logs import (
    get_brace_style_log_with_null_handler,
    main_only_quicksetup_rootlogger,
)

TESTING_VERBOSE = True
TESTING_USE_PRETTY_LOGS = True  # False to make this standalone

DEBUG_INTERNALS = False

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Helper functions
# =============================================================================

def repr_parameter(param: inspect.Parameter) -> str:
    """
    Provides a ``repr``-style representation of a function parameter.
    """
    return (
        f"Parameter(name={param.name}, annotation={param.annotation}, "
        f"kind={param.kind}, default={param.default}"
    )


def get_namespace(fn: Callable, namespace: Optional[str]) -> str:
    """
    Returns a representation of a function's name (perhaps within a namespace),
    like

    .. code-block:: none

        mymodule:MyClass.myclassfunc  # with no namespace
        mymodule:MyClass.myclassfunc|somenamespace  # with a namespace

    Args:
        fn: a function
        namespace: an optional namespace, which can be of any type but is
            normally a ``str``; if not ``None``, ``str(namespace)`` will be
            added to the result. See
            https://dogpilecache.readthedocs.io/en/latest/api.html#dogpile.cache.region.CacheRegion.cache_on_arguments
    """  # noqa
    # See hidden attributes with dir(fn)
    # noinspection PyUnresolvedReferences
    return "{module}:{name}{extra}".format(
        module=fn.__module__,
        name=fn.__qualname__,  # __qualname__ includes class name, if present
        extra=f"|{namespace}" if namespace is not None else "",
    )


# =============================================================================
# New function key generators
# =============================================================================

def fkg_allowing_type_hints(
        namespace: Optional[str],
        fn: Callable,
        to_str: Callable[[Any], str] = repr) -> Callable[[Any], str]:
    """
    Replacement for :func:`dogpile.cache.util.function_key_generator` that
    handles type-hinted functions like

    .. code-block:: python

        def testfunc(param: str) -> str:
            return param + "hello"

    ... at which :func:`inspect.getargspec` balks; plus
    :func:`inspect.getargspec` is deprecated in Python 3.

    Used as an argument to e.g. ``@cache_region_static.cache_on_arguments()``.

    Also modified to make the cached function unique per INSTANCE for normal
    methods of a class.

    Args:
        namespace: optional namespace, as per :func:`get_namespace`
        fn: function to generate a key for (usually the function being
            decorated)
        to_str: function to apply to map arguments to a string (to make a
            unique key for a particular call to the function); by default it
            is :func:`repr`

    Returns:
        a function that generates a string key, based on a given function as
        well as arguments to the returned function itself.
    """

    namespace = get_namespace(fn, namespace)

    sig = inspect.signature(fn)
    argnames = [p.name for p in sig.parameters.values()
                if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD]
    has_self = bool(argnames and argnames[0] in ('self', 'cls'))

    def generate_key(*args: Any, **kw: Any) -> str:
        """
        Makes the actual key for a specific call to the decorated function,
        with particular ``args``/``kwargs``.
        """
        if kw:
            raise ValueError("This dogpile.cache key function generator, "
                             "fkg_allowing_type_hints, "
                             "does not accept keyword arguments.")
        if has_self:
            # Unlike dogpile's default, make it instance- (or class-) specific
            # by including a representation of the "self" or "cls" argument:
            args = [hex(id(args[0]))] + list(args[1:])
        key = namespace + "|" + " ".join(map(to_str, args))
        if DEBUG_INTERNALS:
            log.debug("fkg_allowing_type_hints.generate_key() -> {!r}", key)
        return key

    return generate_key


def multikey_fkg_allowing_type_hints(
        namespace: Optional[str],
        fn: Callable,
        to_str: Callable[[Any], str] = repr) -> Callable[[Any], List[str]]:
    """
    Equivalent of :func:`dogpile.cache.util.function_multi_key_generator`, but
    using :func:`inspect.signature` instead.

    Also modified to make the cached function unique per INSTANCE for normal
    methods of a class.
    """

    namespace = get_namespace(fn, namespace)

    sig = inspect.signature(fn)
    argnames = [p.name for p in sig.parameters.values()
                if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD]
    has_self = bool(argnames and argnames[0] in ('self', 'cls'))

    def generate_keys(*args: Any, **kw: Any) -> List[str]:
        if kw:
            raise ValueError("This dogpile.cache key function generator, "
                             "multikey_fkg_allowing_type_hints, "
                             "does not accept keyword arguments.")
        if has_self:
            # Unlike dogpile's default, make it instance- (or class-) specific
            # by including a representation of the "self" or "cls" argument:
            args = [hex(id(args[0]))] + list(args[1:])
        keys = [namespace + "|" + key for key in map(to_str, args)]
        if DEBUG_INTERNALS:
            log.debug(
                "multikey_fkg_allowing_type_hints.generate_keys() -> {!r}",
                keys)
        return keys

    return generate_keys


def kw_fkg_allowing_type_hints(
        namespace: Optional[str],
        fn: Callable,
        to_str: Callable[[Any], str] = repr) -> Callable[[Any], str]:
    """
    As for :func:`fkg_allowing_type_hints`, but allowing keyword arguments.

    For ``kwargs`` passed in, we will build a ``dict`` of all argname (key) to
    argvalue (values) pairs, including default args from the argspec, and then
    alphabetize the list before generating the key.

    NOTE ALSO that once we have keyword arguments, we should be using
    :func:`repr`, because we need to distinguish

    .. code-block:: python

        kwargs = {'p': 'another', 'q': 'thing'}
        # ... which compat.string_type will make into
        #         p=another q=thing
        # ... from
        kwargs = {'p': 'another q=thing'}

    Also modified to make the cached function unique per INSTANCE for normal
    methods of a class.
    """

    namespace = get_namespace(fn, namespace)

    sig = inspect.signature(fn)
    parameters = list(sig.parameters.values())  # convert from odict_values
    argnames = [p.name for p in parameters
                if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD]
    has_self = bool(argnames and argnames[0] in ('self', 'cls'))

    if DEBUG_INTERNALS:
        log.debug(
            "At start of kw_fkg_allowing_type_hints: namespace={namespace},"
            "parameters=[{parameters}], argnames={argnames}, "
            "has_self={has_self}, fn={fn}",
            namespace=namespace,
            parameters=", ".join(repr_parameter(p) for p in parameters),
            argnames=repr(argnames),
            has_self=has_self,
            fn=repr(fn),
        )

    def generate_key(*args: Any, **kwargs: Any) -> str:
        as_kwargs = {}  # type: Dict[str, Any]
        loose_args = []  # type: List[Any]  # those captured by *args
        # 1. args: get the name as well.
        for idx, arg in enumerate(args):
            if idx >= len(argnames):
                # positional argument to be scooped up with *args
                loose_args.append(arg)
            else:
                # normal plain positional argument
                if has_self and idx == 0:  # "self" or "cls" initial argument
                    argvalue = hex(id(arg))
                else:
                    argvalue = arg
                as_kwargs[argnames[idx]] = argvalue
        # 1b. args with no name
        if loose_args:
            as_kwargs['*args'] = loose_args
            # '*args' is guaranteed not to be a parameter name in its own right
        # 2. kwargs
        as_kwargs.update(kwargs)
        # 3. default values
        for param in parameters:
            if param.default != inspect.Parameter.empty:
                if param.name not in as_kwargs:
                    as_kwargs[param.name] = param.default
        # 4. sorted by name
        #    ... but also incorporating the name of the argument, because once
        #    we allow the arbitrary **kwargs format, order is no longer
        #    sufficient to discriminate
        #       fn(p="another", q="thing")
        #    from
        #       fn(r="another", s="thing")
        argument_values = ["{k}={v}".format(k=key, v=to_str(as_kwargs[key]))
                           for key in sorted(as_kwargs.keys())]
        key = namespace + '|' + " ".join(argument_values)
        if DEBUG_INTERNALS:
            log.debug("kw_fkg_allowing_type_hints.generate_key() -> {!r}", key)
        return key

    return generate_key


# =============================================================================
# Default function key generator with a short name
# =============================================================================

fkg = kw_fkg_allowing_type_hints

# Can now do:
#
# @mycache.cache_on_arguments(function_key_generator=fkg)
# def myfunc():
#     pass


# =============================================================================
# Unit tests
# =============================================================================

def unit_test_cache() -> None:
    mycache = make_region()
    mycache.configure(backend='dogpile.cache.memory')

    plain_fkg = fkg_allowing_type_hints
    kw_fkg = kw_fkg_allowing_type_hints
    # I'm not sure what dogpile.cache.utils.function_multi_key_generator is
    # used for, so haven't fully tested multikey_fkg_allowing_type_hints, but
    # it works internally in the same way as fkg_allowing_type_hints.

    fn_was_called = False

    def test(result: str, should_call_fn: bool, reset: bool = True) -> None:
        nonlocal fn_was_called
        log.info("{}", result)
        assert fn_was_called == should_call_fn, (
            f"fn_was_called={fn_was_called}, should_call_fn={should_call_fn}")
        if reset:
            fn_was_called = False

    def fn_called(text: str) -> None:
        log.info("{}", text)
        nonlocal fn_was_called
        fn_was_called = True

    @mycache.cache_on_arguments(function_key_generator=None)
    def no_params_dogpile_default_fkg():  # no type hints!
        fn_called("CACHED FUNCTION no_params_dogpile_default_fkg() CALLED")
        return "no_params_dogpile_default_fkg: hello"

    @mycache.cache_on_arguments(function_key_generator=plain_fkg)
    def noparams() -> str:
        fn_called("CACHED FUNCTION noparams() CALLED")
        return "noparams: hello"

    @mycache.cache_on_arguments(function_key_generator=plain_fkg)
    def oneparam(a: str) -> str:
        fn_called("CACHED FUNCTION oneparam() CALLED")
        return "oneparam: hello, " + a

    @mycache.cache_on_arguments(function_key_generator=plain_fkg)
    def twoparam_with_default_wrong_dec(a: str, b: str = "Zelda") -> str:
        fn_called("CACHED FUNCTION twoparam_with_default_wrong_dec() CALLED")
        return ("twoparam_with_default_wrong_dec: hello, " + a +
                "; this is " + b)

    @mycache.cache_on_arguments(function_key_generator=kw_fkg)
    def twoparam_with_default_right_dec(a: str, b: str = "Zelda") -> str:
        fn_called("CACHED FUNCTION twoparam_with_default_right_dec() CALLED")
        return ("twoparam_with_default_right_dec: hello, " + a +
                "; this is " + b)

    @mycache.cache_on_arguments(function_key_generator=kw_fkg)
    def twoparam_all_defaults_no_typehints(a="David", b="Zelda"):
        fn_called("CACHED FUNCTION twoparam_all_defaults_no_typehints() "
                  "CALLED")
        return ("twoparam_all_defaults_no_typehints: hello, " + a +
                "; this is " + b)

    @mycache.cache_on_arguments(function_key_generator=kw_fkg)
    def fn_args_kwargs(*args, **kwargs):
        fn_called("CACHED FUNCTION fn_args_kwargs() CALLED")
        return f"fn_args_kwargs: args={args!r}, kwargs={kwargs!r}"

    @mycache.cache_on_arguments(function_key_generator=kw_fkg)
    def fn_all_possible(a, b, *args, d="David", **kwargs):
        fn_called("CACHED FUNCTION fn_all_possible() CALLED")
        return (
            f"fn_all_possible: a={a!r}, b={b!r}, args={args!r}, d={d!r}, "
            f"kwargs={kwargs!r}"
        )

    class TestClass(object):
        c = 999

        def __init__(self, a: int = 200) -> None:
            self.a = a

        @mycache.cache_on_arguments(function_key_generator=None)
        def no_params_dogpile_default_fkg(self):  # no type hints!
            fn_called("CACHED FUNCTION TestClass."
                      "no_params_dogpile_default_fkg() CALLED")
            return "TestClass.no_params_dogpile_default_fkg: hello"

        @mycache.cache_on_arguments(function_key_generator=None)
        def dogpile_default_test_2(self):  # no type hints!
            fn_called("CACHED FUNCTION TestClass."
                      "dogpile_default_test_2() CALLED")
            return "TestClass.dogpile_default_test_2: hello"

        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def noparams(self) -> str:
            fn_called("CACHED FUNCTION TestClass.noparams() CALLED")
            return f"Testclass.noparams: hello; a={self.a}"

        @mycache.cache_on_arguments(function_key_generator=kw_fkg)
        def no_params_instance_cache(self) -> str:
            fn_called("PER-INSTANCE-CACHED FUNCTION "
                      "TestClass.no_params_instance_cache() CALLED")
            return f"TestClass.no_params_instance_cache: a={self.a}"

        # Decorator order is critical here:
        # https://stackoverflow.com/questions/1987919/why-can-decorator-not-decorate-a-staticmethod-or-a-classmethod  # noqa
        @classmethod
        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def classy(cls) -> str:
            fn_called("CACHED FUNCTION TestClass.classy() CALLED")
            return f"TestClass.classy: hello; c={cls.c}"

        @staticmethod
        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def static() -> str:
            fn_called("CACHED FUNCTION TestClass.static() CALLED")
            return "TestClass.static: hello"

        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def oneparam(self, q: str) -> str:
            fn_called("CACHED FUNCTION TestClass.oneparam() CALLED")
            return "TestClass.oneparam: hello, " + q

        @mycache.cache_on_arguments(function_key_generator=kw_fkg)
        def fn_all_possible(self, a, b, *args, d="David", **kwargs):
            fn_called("CACHED FUNCTION TestClass.fn_all_possible() CALLED")
            return (
                f"TestClass.fn_all_possible: a={a!r}, b={b!r}, args={args!r}, "
                f"d={d!r}, kwargs={kwargs!r}")

        @property
        @mycache.cache_on_arguments(function_key_generator=kw_fkg)
        def prop(self) -> str:
            fn_called("CACHED PROPERTY TestClass.prop() CALLED")
            return f"TestClass.prop: a={self.a!r}"

    class SecondTestClass:
        def __init__(self) -> None:
            self.d = 5

        @mycache.cache_on_arguments(function_key_generator=None)
        def dogpile_default_test_2(self):  # no type hints!
            fn_called("CACHED FUNCTION SecondTestClass."
                      "dogpile_default_test_2() CALLED")
            return "SecondTestClass.dogpile_default_test_2: hello"

    class Inherited(TestClass):
        def __init__(self, a=101010):
            super().__init__(a=a)

        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def noparams(self) -> str:
            fn_called("CACHED FUNCTION Inherited.noparams() CALLED")
            return f"Inherited.noparams: hello; a={self.a}"

        @mycache.cache_on_arguments(function_key_generator=kw_fkg)
        def no_params_instance_cache(self) -> str:
            fn_called("PER-INSTANCE-CACHED FUNCTION "
                      "Inherited.no_params_instance_cache() CALLED")
            return f"Inherited.no_params_instance_cache: a={self.a}"

        # Decorator order is critical here:
        # https://stackoverflow.com/questions/1987919/why-can-decorator-not-decorate-a-staticmethod-or-a-classmethod  # noqa
        @classmethod
        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def classy(cls) -> str:
            fn_called("CACHED FUNCTION Inherited.classy() CALLED")
            return f"Inherited.classy: hello; c={cls.c}"

        @staticmethod
        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def static() -> str:
            fn_called("CACHED FUNCTION Inherited.static() CALLED")
            return "Inherited.static: hello"

        @mycache.cache_on_arguments(function_key_generator=plain_fkg)
        def oneparam(self, q: str) -> str:
            fn_called("CACHED FUNCTION Inherited.oneparam() CALLED")
            return "Inherited.oneparam: hello, " + q

        # BUT fn_all_possible IS NOT OVERRIDDEN

        @property
        @mycache.cache_on_arguments(function_key_generator=kw_fkg)
        def prop(self) -> str:
            fn_called("CACHED PROPERTY Inherited.prop() CALLED")
            return f"Inherited.prop: a={self.a!r}"

    log.warning("Fetching cached information #1 (should call noparams())...")
    test(noparams(), True)
    log.warning("Fetching cached information #2 (should not call noparams())...")  # noqa
    test(noparams(), False)

    log.warning("Testing functions with other signatures...")
    test(oneparam("Arthur"), True)
    test(oneparam("Arthur"), False)
    test(oneparam("Bob"), True)
    test(oneparam("Bob"), False)
    test(twoparam_with_default_wrong_dec("Celia"), True)
    test(twoparam_with_default_wrong_dec("Celia"), False)
    test(twoparam_with_default_wrong_dec("Celia", "Yorick"), True)
    test(twoparam_with_default_wrong_dec("Celia", "Yorick"), False)

    log.warning("Trying with keyword arguments and wrong key generator")
    try:
        log.info("{}", twoparam_with_default_wrong_dec(a="Celia", b="Yorick"))
        raise AssertionError("Inappropriate success with keyword arguments!")
    except ValueError:
        log.info("Correct rejection of keyword arguments")

    log.warning("Trying with keyword arguments and right key generator")
    test(twoparam_with_default_right_dec(a="Celia"), True)
    test(twoparam_with_default_right_dec(a="Celia", b="Yorick"), True)
    test(twoparam_with_default_right_dec(b="Yorick", a="Celia"), False)
    test(twoparam_with_default_right_dec("Celia", b="Yorick"), False)

    test(twoparam_all_defaults_no_typehints(), True)
    test(twoparam_all_defaults_no_typehints(a="Edith"), True)
    test(twoparam_all_defaults_no_typehints(a="Edith"), False)
    test(twoparam_all_defaults_no_typehints(b="Romeo"), True)
    test(twoparam_all_defaults_no_typehints(b="Romeo"), False)
    test(twoparam_all_defaults_no_typehints("Greta", b="Romeo"), True)
    test(twoparam_all_defaults_no_typehints("Greta", b="Romeo"), False)
    test(twoparam_all_defaults_no_typehints(a="Felicity", b="Sigurd"), True)
    test(twoparam_all_defaults_no_typehints(a="Felicity", b="Sigurd"), False)
    test(twoparam_all_defaults_no_typehints("Felicity", "Sigurd"), False)
    test(twoparam_all_defaults_no_typehints("Felicity", "Sigurd"), False)
    test(twoparam_all_defaults_no_typehints(b="Sigurd", a="Felicity"), False)
    test(twoparam_all_defaults_no_typehints(b="Sigurd", a="Felicity"), False)

    test(fn_args_kwargs(1, 2, 3, d="David", f="Edgar"), True)
    test(fn_args_kwargs(1, 2, 3, d="David", f="Edgar"), False)

    test(fn_args_kwargs(p="another", q="thing"), True)
    test(fn_args_kwargs(p="another", q="thing"), False)
    log.warning("The next call MUST NOT go via the cache, i.e. func should be CALLED")  # noqa
    test(fn_args_kwargs(p="another q=thing"), True)
    test(fn_args_kwargs(p="another q=thing"), False)

    test(fn_all_possible(10, 11, 12, "Horace", "Iris"), True)
    test(fn_all_possible(10, 11, 12, "Horace", "Iris"), False)
    test(fn_all_possible(10, 11, 12, d="Horace"), True)
    test(fn_all_possible(10, 11, 12, d="Horace"), False)
    test(fn_all_possible(98, 99, d="Horace"), True)
    test(fn_all_possible(98, 99, d="Horace"), False)
    test(fn_all_possible(98, 99, d="Horace", p="another", q="thing"), True)
    test(fn_all_possible(98, 99, d="Horace", p="another", q="thing"), False)
    test(fn_all_possible(98, 99, d="Horace", r="another", s="thing"), True)
    test(fn_all_possible(98, 99, d="Horace", r="another", s="thing"), False)

    log.warning("Testing class member functions")
    t = TestClass()
    test(t.noparams(), True)
    test(t.noparams(), False)
    test(t.classy(), True)
    test(t.classy(), False)
    test(t.static(), True)
    test(t.static(), False)
    test(t.oneparam("Arthur"), True)
    test(t.oneparam("Arthur"), False)
    test(t.oneparam("Bob"), True)
    test(t.oneparam("Bob"), False)
    test(t.fn_all_possible(10, 11, 12, "Horace", "Iris"), True)
    test(t.fn_all_possible(10, 11, 12, "Horace", "Iris"), False)
    test(t.fn_all_possible(10, 11, 12, d="Horace"), True)
    test(t.fn_all_possible(10, 11, 12, d="Horace"), False)
    test(t.fn_all_possible(98, 99, d="Horace"), True)
    test(t.fn_all_possible(98, 99, d="Horace"), False)
    test(t.fn_all_possible(98, 99, d="Horace", p="another", q="thing"), True)
    test(t.fn_all_possible(98, 99, d="Horace", p="another", q="thing"), False)
    test(t.fn_all_possible(98, 99, d="Horace", r="another", s="thing"), True)
    test(t.fn_all_possible(98, 99, d="Horace", r="another", s="thing"), False)
    test(t.prop, True)
    test(t.prop, False)

    log.warning("Testing functions for another INSTANCE of the same class")
    t_other = TestClass(a=999)
    test(t_other.noparams(), True)
    test(t_other.noparams(), False)
    test(t_other.classy(), False)  # SAME CLASS as t; shouldn't be re-called
    test(t_other.classy(), False)
    test(t_other.static(), False)  # SAME CLASS as t; shouldn't be re-called
    test(t_other.static(), False)
    test(t_other.oneparam("Arthur"), True)
    test(t_other.oneparam("Arthur"), False)
    test(t_other.oneparam("Bob"), True)
    test(t_other.oneparam("Bob"), False)
    test(t_other.fn_all_possible(10, 11, 12, "Horace", "Iris"), True)
    test(t_other.fn_all_possible(10, 11, 12, "Horace", "Iris"), False)
    test(t_other.fn_all_possible(10, 11, 12, d="Horace"), True)
    test(t_other.fn_all_possible(10, 11, 12, d="Horace"), False)
    test(t_other.fn_all_possible(98, 99, d="Horace"), True)
    test(t_other.fn_all_possible(98, 99, d="Horace"), False)
    test(t_other.fn_all_possible(98, 99, d="Horace", p="another", q="thing"), True)  # noqa
    test(t_other.fn_all_possible(98, 99, d="Horace", p="another", q="thing"), False)  # noqa
    test(t_other.fn_all_possible(98, 99, d="Horace", r="another", s="thing"), True)  # noqa
    test(t_other.fn_all_possible(98, 99, d="Horace", r="another", s="thing"), False)  # noqa
    test(t_other.prop, True)
    test(t_other.prop, False)

    test(t.no_params_instance_cache(), True)
    test(t.no_params_instance_cache(), False)
    test(t_other.no_params_instance_cache(), True)
    test(t_other.no_params_instance_cache(), False)

    log.warning("Testing functions for instance of a derived class")
    t_inh = Inherited(a=777)
    test(t_inh.noparams(), True)
    test(t_inh.noparams(), False)
    test(t_inh.classy(), True)
    test(t_inh.classy(), False)
    test(t_inh.static(), True)
    test(t_inh.static(), False)
    test(t_inh.oneparam("Arthur"), True)
    test(t_inh.oneparam("Arthur"), False)
    test(t_inh.oneparam("Bob"), True)
    test(t_inh.oneparam("Bob"), False)
    test(t_inh.fn_all_possible(10, 11, 12, "Horace", "Iris"), True)
    test(t_inh.fn_all_possible(10, 11, 12, "Horace", "Iris"), False)
    test(t_inh.fn_all_possible(10, 11, 12, d="Horace"), True)
    test(t_inh.fn_all_possible(10, 11, 12, d="Horace"), False)
    test(t_inh.fn_all_possible(98, 99, d="Horace"), True)
    test(t_inh.fn_all_possible(98, 99, d="Horace"), False)
    test(t_inh.fn_all_possible(98, 99, d="Horace", p="another", q="thing"), True)  # noqa
    test(t_inh.fn_all_possible(98, 99, d="Horace", p="another", q="thing"), False)  # noqa
    test(t_inh.fn_all_possible(98, 99, d="Horace", r="another", s="thing"), True)  # noqa
    test(t_inh.fn_all_possible(98, 99, d="Horace", r="another", s="thing"), False)  # noqa
    test(t_inh.prop, True)
    test(t_inh.prop, False)

    test(no_params_dogpile_default_fkg(), True)
    test(no_params_dogpile_default_fkg(), False)
    try:
        test(t.no_params_dogpile_default_fkg(), True)
        log.info("dogpile.cache default FKG correctly distinguishing between "
                 "plain and class-member function in same module")
    except AssertionError:
        log.warning("Known dogpile.cache default FKG problem - conflates "
                    "plain/class member function of same name (unless "
                    "namespace is manually given)")
    test(t.no_params_dogpile_default_fkg(), False)

    t2 = SecondTestClass()
    test(t.dogpile_default_test_2(), True)
    test(t.dogpile_default_test_2(), False)
    try:
        test(t2.dogpile_default_test_2(), True)
        log.info("dogpile.cache default FKG correctly distinguishing between "
                 "member functions of two different classes with same name")
    except AssertionError:
        log.warning("Known dogpile.cache default FKG problem - conflates "
                    "member functions of two different classes where "
                    "functions have same name (unless namespace is manually "
                    "given)")
    test(t2.dogpile_default_test_2(), False)

    log.info("Success!")


# TEST THIS WITH:
# python -m cardinal_pythonlib.dogpile_cache
if __name__ == '__main__':
    level = logging.DEBUG if TESTING_VERBOSE else logging.INFO
    if TESTING_USE_PRETTY_LOGS:
        main_only_quicksetup_rootlogger(level=level)
    else:
        logging.basicConfig(level=level)
    unit_test_cache()



#!/usr/bin/env python
# cardinal_pythonlib/psychiatry/rfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**WORK IN PROGRESS. Doesn't do much that's useful at present.**

Functions to be used from R via reticulate
(https://cran.r-project.org/web/packages/reticulate/index.html).

See ``drugs.py`` for notes on how to get reticulate talking to this library.

Briefly:

.. code-block:: r

    # -------------------------------------------------------------------------
    # Load libraries
    # -------------------------------------------------------------------------

    library(data.table)
    library(reticulate)

    # -------------------------------------------------------------------------
    # Set up reticulate
    # -------------------------------------------------------------------------

    VENV <- "~/dev/venvs/cardinal_pythonlib"  # or your preferred virtualenv
    PYTHON_EXECUTABLE <- ifelse(
        .Platform$OS.type == "windows",
        file.path(VENV, "Scripts", "python.exe"),  # Windows
        file.path(VENV, "bin", "python")  # Linux
    )
    reticulate::use_python(PYTHON_EXECUTABLE, required=TRUE)

    # -------------------------------------------------------------------------
    # Import Python modules
    # -------------------------------------------------------------------------

    cpl_rfunc <- reticulate::import("cardinal_pythonlib.psychiatry.rfunc")

    # -------------------------------------------------------------------------
    # Try other things
    # -------------------------------------------------------------------------

    repl_python()  # start an interactive Python session

"""  # noqa

import sys
from typing import Any, Dict


# =============================================================================
# Simple information for R
# =============================================================================

def get_python_repr(x: Any) -> str:
    r"""
    A few notes:

    **data.table()**

    Data tables are converted to a Python dictionary:

    .. code-block:: r

        dt <- data.table(
            subject = c("Alice", "Bob", "Charles", "Dawn", "Egbert", "Flora"),
            drug = c("citalopram", "Cipramil", "Prozac", "fluoxetine",
                     "Priadel", "Haldol")
        )
        dt_repr <- cpl_rfunc$get_python_repr(dt)

    gives this when a duff import is used via
    ``reticulate::import_from_path()``:

    .. code-block:: none

        [1] "{'drug': ['citalopram', 'Cipramil', 'Prozac', 'fluoxetine',
        'Priadel', 'Haldol'], 'subject': ['Alice', 'Bob', 'Charles', 'Dawn',
        'Egbert', 'Flora']}"

    but this when a proper import is used via ``reticulate::import()``:

    .. code-block:: none

        [1] "   subject        drug\n0    Alice  citalopram\n1      Bob
        Cipramil\n2  Charles      Prozac\n3     Dawn  fluoxetine\n4   Egbert
         Priadel\n5    Flora      Haldol"

    """
    return repr(x)


def get_python_repr_of_type(x: Any) -> str:
    """
    See ``get_python_repr``.

    .. code-block:: r

        dt_type_repr <- cpl_rfunc$get_python_repr_of_type(dt)

    gives this when a duff import is used via
    ``reticulate::import_from_path()``:

    .. code-block:: none

        [1] "<class 'dict'>"

    but this when a proper import is used via ``reticulate::import()``:

    .. code-block:: none

        [1] "<class 'pandas.core.frame.DataFrame'>"

    The same is true of a data.frame().

    """
    return repr(type(x))


def test_get_dict() -> Dict[str, Any]:
    """
    Test with:

    .. code-block:: r

        testdict <- cpl_rfunc$test_get_dict()

    This gives a list:

    .. code-block:: none

        > testdict
        $strings
        [1] "one"   "two"   "three" "four"  "five"

        $integers
        [1] 1 2 3 4 5

        $floats
        [1] 1.1 2.1 3.1 4.1 5.1

    """
    return {
        'integers': [1, 2, 3, 4, 5],
        'floats': [1.1, 2.1, 3.1, 4.1, 5.1],
        'strings': ["one", "two", "three", "four", "five"],
    }


def flush_stdout_stderr() -> None:
    """
    R code won't see much unless we flush stdout/stderr manually.
    See also https://github.com/rstudio/reticulate/issues/284
    """
    sys.stdout.flush()
    sys.stderr.flush()



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/sqlserver.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**SQLAlchemy functions specific to Microsoft SQL Server.**

"""

from contextlib import contextmanager

from sqlalchemy.orm import Session as SqlASession

from cardinal_pythonlib.sqlalchemy.dialect import quote_identifier
from cardinal_pythonlib.sqlalchemy.engine_func import is_sqlserver
from cardinal_pythonlib.sqlalchemy.session import get_engine_from_session


# =============================================================================
# Workarounds for SQL Server "DELETE takes forever" bug
# =============================================================================

@contextmanager
def if_sqlserver_disable_constraints(session: SqlASession,
                                     tablename: str) -> None:
    """
    If we're running under SQL Server, disable constraint checking for the
    specified table while the resource is held.

    Args:
        session: SQLAlchemy :class:`Session`
        tablename: table name

    See
    https://stackoverflow.com/questions/123558/sql-server-2005-t-sql-to-temporarily-disable-a-trigger
    """  # noqa
    engine = get_engine_from_session(session)
    if is_sqlserver(engine):
        quoted_tablename = quote_identifier(tablename, engine)
        session.execute(
            f"ALTER TABLE {quoted_tablename} NOCHECK CONSTRAINT all")
        yield
        session.execute(
            f"ALTER TABLE {quoted_tablename} WITH CHECK CHECK CONSTRAINT all")
    else:
        yield


@contextmanager
def if_sqlserver_disable_triggers(session: SqlASession,
                                  tablename: str) -> None:
    """
    If we're running under SQL Server, disable triggers for the specified table
    while the resource is held.

    Args:
        session: SQLAlchemy :class:`Session`
        tablename: table name

    See
    https://stackoverflow.com/questions/123558/sql-server-2005-t-sql-to-temporarily-disable-a-trigger
    """  # noqa
    engine = get_engine_from_session(session)
    if is_sqlserver(engine):
        quoted_tablename = quote_identifier(tablename, engine)
        session.execute(
            f"ALTER TABLE {quoted_tablename} DISABLE TRIGGER all")
        yield
        session.execute(
            f"ALTER TABLE {quoted_tablename} ENABLE TRIGGER all")
    else:
        yield


@contextmanager
def if_sqlserver_disable_constraints_triggers(session: SqlASession,
                                              tablename: str) -> None:
    """
    If we're running under SQL Server, disable triggers AND constraints for the
    specified table while the resource is held.

    Args:
        session: SQLAlchemy :class:`Session`
        tablename: table name
    """
    with if_sqlserver_disable_constraints(session, tablename):
        with if_sqlserver_disable_triggers(session, tablename):
            yield



#!/usr/bin/env python
# cardinal_pythonlib/module_version.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Find a Python module's version (across Python versions)**

See

- https://stackoverflow.com/questions/710609/checking-a-python-module-version-at-runtime
- https://stackoverflow.com/questions/20180543/how-to-check-version-of-python-modules

"""  # noqa

# =============================================================================
# Imports
# =============================================================================

import sys

from semantic_version import Version

# noinspection PyUnresolvedReferences
import cardinal_pythonlib.ensure_test_executed_correctly

if sys.version_info > (3, 8):
    # Python 3.8 or higher
    # noinspection PyCompatibility
    from importlib.metadata import version
else:
    try:
        from importlib_metadata import version
    except ImportError:
        raise


# =============================================================================
# Report Python module versions
# =============================================================================

def module_version_string(module_name: str) -> str:
    """
    The string version of a Python module.

    Will raise :exc:`PackageNotFoundError` if not found.
    """
    return version(module_name)


def module_semantic_version(module_name: str) -> Version:
    """
    The semantic version of a Python module.
    """
    version_string = module_version_string(module_name)  # may raise PackageNotFoundError  # noqa
    return Version(version_string)  # may raise ValueError



#!/usr/bin/env python
# cardinal_pythonlib/rate_limiting.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Rate-limiting functions.**

"""

import logging
from time import perf_counter, sleep
from typing import Any, Callable, Optional, Union

from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger


log = logging.getLogger(__name__)
FuncType = Callable[..., Any]


def rate_limited(max_per_second: Optional[Union[int, float]]) \
        -> Callable[[FuncType], FuncType]:
    """
    Returns a function that rate-limits another function to the specified
    frequency. Can be used as a decorator, e.g.

    .. code-block:: python

        from cardinal_pythonlib.rate_limiting import rate_limited

        @rate_limited(2)
        def do_something_max_2hz():
            print("tick...")

        for i in range(10):
            do_something_max_2hz()

    or dynamically:

    .. code-block:: python

        from cardinal_pythonlib.rate_limiting import rate_limited

        def do_something():
            print("tick...")

        limited = rate_limited(2)(do_something)
        for i in range(10):
            limited()

    Args:
        max_per_second:
            maximum number of calls per second to permit, or ``None`` for no
            limit

    Returns:
        a function that takes a function argument

    Based on
    https://www.gregburek.com/2011/12/05/rate-limiting-with-decorators/, with
    minor modifications.
    """
    assert max_per_second is None or max_per_second > 0
    min_interval = None  # type: Optional[float]
    if max_per_second is not None:
        min_interval = 1.0 / float(max_per_second)

    def decorate(func: FuncType) -> FuncType:
        if max_per_second is None:
            return func  # no rate limiting
        last_time_called = 0.0

        def rate_limited_function(*args, **kwargs) -> Any:
            nonlocal last_time_called
            elapsed = perf_counter() - last_time_called
            left_to_wait = min_interval - elapsed
            if left_to_wait > 0:
                sleep(left_to_wait)
            retval = func(*args, **kwargs)
            last_time_called = perf_counter()
            return retval

        return rate_limited_function

    return decorate


@rate_limited(2)
def _test_print_2hz(num: int) -> None:
    log.info(f"_test_print_2hz: {num}")


@rate_limited(5)
def _test_print_5hz(num: int) -> None:
    log.info(f"_test_print_5hz: {num}")


def _test_print(num: int) -> None:
    log.info(f"_test_print: {num}")


def test_rate_limiter() -> None:
    """
    Test the rate-limiting functions.
    """
    n = 10
    log.warning("Via decorator, 2 Hz")
    for i in range(1, n + 1):
        _test_print_2hz(i)
    log.warning("Via decorator, 5 Hz")
    for i in range(1, n + 1):
        _test_print_5hz(i)
    log.warning("Created dynamically, 10 Hz")
    tenhz = rate_limited(10)(_test_print)
    for i in range(1, n + 1):
        tenhz(i)
    log.warning("Created dynamically, unlimited")
    unlimited = rate_limited(None)(_test_print)
    for i in range(1, n + 1):
        unlimited(i)


if __name__ == "__main__":
    main_only_quicksetup_rootlogger(level=logging.DEBUG)
    test_rate_limiter()



#!/usr/bin/env python
# cardinal_pythonlib/iterhelp.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Iteration assistance functions.**

"""

from itertools import product
from typing import Dict, Iterable


def product_dict(**kwargs: Iterable) -> Iterable[Dict]:
    """
    See
    https://stackoverflow.com/questions/5228158/cartesian-product-of-a-dictionary-of-lists.

    Takes keyword arguments, and yields dictionaries containing every
    combination of possibilities for each keyword.

    Examples:

    .. code-block:: python

        >>> list(product_dict(a=[1, 2], b=[3, 4]))
        [{'a': 1, 'b': 3}, {'a': 1, 'b': 4}, {'a': 2, 'b': 3}, {'a': 2, 'b': 4}]

        >>> list(product_dict(a="x", b=range(3)))
        [{'a': 'x', 'b': 0}, {'a': 'x', 'b': 1}, {'a': 'x', 'b': 2}]

        >>> product_dict(a="x", b=range(3))
        <generator object product_dict at 0x7fb328070678>
    """  # noqa
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in product(*vals):
        yield dict(zip(keys, instance))



#!/usr/bin/env python
# cardinal_pythonlib/pyramid/responses.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Specialized response types for Pyramid (which implement MIME types and
suggested download methods, etc.).**

"""

# noinspection PyUnresolvedReferences
from pyramid.response import Response

from cardinal_pythonlib.httpconst import MimeType


# =============================================================================
# Character sets
# =============================================================================

UTF8 = "UTF-8"


# =============================================================================
# Responses
# =============================================================================

class BinaryResponse(Response):
    """
    Base class for creating binary HTTP responses.
    """
    def __init__(self, body: bytes, filename: str,
                 content_type: str, as_inline: bool = False, **kwargs) -> None:
        """
        Args:
            body: binary data
            filename: filename to associate with the download
            content_type: MIME content type
            as_inline: inline, rather than as an attachment?
                Inline: display within browser, if possible.
                Attachment: download.
        """
        disp = "inline" if as_inline else "attachment"
        super().__init__(
            content_type=content_type,
            content_disposition=f"{disp}; filename={filename}",
            content_encoding="binary",
            content_length=len(body),
            body=body,
            **kwargs
        )


class JsonAttachmentResponse(Response):
    """
    Response class for returning a JSON file to the user as an attachment.
    """
    def __init__(self, body: str, filename: str, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.JSON,
            content_disposition=f"attachment; filename={filename}",
            body=body,
            charset=UTF8,
            **kwargs
        )


class JsonResponse(Response):
    """
    Response class for showing JSON in the browser.
    """
    def __init__(self, body: str, **kwargs) -> None:
        # Watch out here.
        # TypeError: You cannot set the body to a text value without a charset
        # https://github.com/Pylons/webob/issues/298
        super().__init__(
            content_type=MimeType.JSON,
            body=body,
            charset=UTF8,
            **kwargs
        )


class OdsResponse(BinaryResponse):
    """
    Response class for returning an ODS (OpenOffice Spreadsheet) file to the
    user.
    """
    def __init__(self, body: bytes, filename: str, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.ODS,
            body=body,
            filename=filename,
            **kwargs
        )


class PdfResponse(BinaryResponse):
    """
    Response class for returning a PDF to the user.
    """
    def __init__(self, body: bytes, filename: str,
                 as_inline: bool = True, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.PDF,
            filename=filename,
            as_inline=as_inline,
            body=body,
            **kwargs
        )


class SqliteBinaryResponse(BinaryResponse):
    """
    Response class for returning a SQLite binary database to the user.
    """
    def __init__(self, body: bytes, filename: str, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.SQLITE3,
            filename=filename,
            body=body,
            **kwargs
        )


class TextAttachmentResponse(Response):
    """
    Response class for returning text to the user as an attachment.
    """
    def __init__(self, body: str, filename: str, **kwargs) -> None:
        # Will default to UTF-8
        super().__init__(
            content_type=MimeType.TEXT,
            content_disposition=f"attachment; filename={filename}",
            body=body,
            **kwargs
        )


class TextResponse(Response):
    """
    Response class for returning text to the user, viewed in the browser.
    """
    def __init__(self, body: str, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.TEXT,
            body=body,
            **kwargs
        )


class TsvResponse(Response):
    """
    Response class for returning a TSV file to the user.
    """
    def __init__(self, body: str, filename: str, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.TSV,
            content_disposition=f"attachment; filename={filename}",
            body=body,
            **kwargs
        )


class XlsxResponse(BinaryResponse):
    """
    Response class for returning an XLSX (Excel) file to the user.
    """
    def __init__(self, body: bytes, filename: str, **kwargs) -> None:
        super().__init__(
            content_type=MimeType.XLSX,
            body=body,
            filename=filename,
            **kwargs
        )


class XmlResponse(Response):
    """
    Response class for returning XML to the user.
    """
    def __init__(self, body: str, **kwargs) -> None:
        # application/xml versus text/xml:
        # https://stackoverflow.com/questions/4832357
        super().__init__(
            content_type=MimeType.XML,
            body=body,
            **kwargs
        )


class ZipResponse(BinaryResponse):
    """
    Response class for returning a ZIP file to the user.
    """
    def __init__(self, body: bytes, filename: str, **kwargs) -> None:
        # For ZIP, "inline" and "attachment" dispositions are equivalent, since
        # browsers don't display ZIP files inline.
        # https://stackoverflow.com/questions/1395151
        super().__init__(
            content_type=MimeType.ZIP,
            filename=filename,
            body=body,
            **kwargs
        )



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/logs.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to assist with SQLAlchemy logs.**

"""

import logging


def pre_disable_sqlalchemy_extra_echo_log() -> None:
    """
    Adds a null handler to the log named ``sqlalchemy.engine.base.Engine``,
    which prevents a duplicated log stream being created later.

    Why is this necessary?

    If you create an SQLAlchemy :class:`Engine` with ``echo=True``, it creates
    an additional log to ``stdout``, via

    .. code-block:: python

        sqlalchemy.engine.base.Engine.__init__()
        sqlalchemy.log.instance_logger()
        sqlalchemy.log.InstanceLogger.__init__()
            # ... which checks that the logger has no handlers and if not calls:
        sqlalchemy.log._add_default_handler()
            # ... which adds a handler to sys.stdout
    """
    log = logging.getLogger("sqlalchemy.engine.base.Engine")
    log.addHandler(logging.NullHandler())



#!/usr/bin/env python
# cardinal_pythonlib/django/reprfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Provide a "repr()"-like function for Django model objects.**

"""

# noinspection PyUnresolvedReferences
from django.core.exceptions import ObjectDoesNotExist


def modelrepr(instance) -> str:
    """
    Default ``repr`` version of a Django model object, for debugging.
    """
    elements = []
    # noinspection PyProtectedMember
    for f in instance._meta.get_fields():
        # https://docs.djangoproject.com/en/2.0/ref/models/meta/
        if f.auto_created:
            continue
        if f.is_relation and f.related_model is None:
            continue
        fieldname = f.name
        try:
            value = repr(getattr(instance, fieldname))
        except ObjectDoesNotExist:
            value = "<RelatedObjectDoesNotExist>"
        elements.append(f"{fieldname}={value}")
    return f"<{type(instance).__name__} <{', '.join(elements)}>>"
    # - type(instance).__name__ gives the Python class name from an instance
    # - ... as does ModelClass.__name__ but we don't have that directly here
    # - instance._meta.model_name gives a lower-case version



#!/usr/bin/env python
# cardinal_pythonlib/nhs.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions regarding NHS numbers, etc.**

"""

import re
import random
from typing import List, Optional, Union

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# NHS number validation
# =============================================================================

NHS_DIGIT_WEIGHTINGS = [10, 9, 8, 7, 6, 5, 4, 3, 2]


def nhs_check_digit(ninedigits: Union[str, List[Union[str, int]]]) -> int:
    """
    Calculates an NHS number check digit.

    Args:
        ninedigits: string or list

    Returns:
        check digit

    Method:

    1. Multiply each of the first nine digits by the corresponding
       digit weighting (see :const:`NHS_DIGIT_WEIGHTINGS`).
    2. Sum the results.
    3. Take remainder after division by 11.
    4. Subtract the remainder from 11
    5. If this is 11, use 0 instead
       If it's 10, the number is invalid
       If it doesn't match the actual check digit, the number is invalid

    """
    if len(ninedigits) != 9 or not all(str(x).isdigit() for x in ninedigits):
        raise ValueError("bad string to nhs_check_digit")
    check_digit = 11 - (sum([
        int(d) * f
        for (d, f) in zip(ninedigits, NHS_DIGIT_WEIGHTINGS)
    ]) % 11)
    # ... % 11 yields something in the range 0-10
    # ... 11 - that yields something in the range 1-11
    if check_digit == 11:
        check_digit = 0
    return check_digit


def is_valid_nhs_number(n: int) -> bool:
    """
    Validates an integer as an NHS number.

    Args:
        n: NHS number

    Returns:
        valid?

    Checksum details are at
    https://www.datadictionary.nhs.uk/version2/data_dictionary/data_field_notes/n/nhs_number_de.asp
    """  # noqa
    if not isinstance(n, int):
        log.debug("is_valid_nhs_number: parameter was not of integer type")
        return False

    s = str(n)
    # Not 10 digits long?
    if len(s) != 10:
        log.debug("is_valid_nhs_number: not 10 digits")
        return False

    main_digits = [int(s[i]) for i in range(9)]
    actual_check_digit = int(s[9])  # tenth digit
    expected_check_digit = nhs_check_digit(main_digits)
    if expected_check_digit == 10:
        log.debug("is_valid_nhs_number: calculated check digit invalid")
        return False
    if expected_check_digit != actual_check_digit:
        log.debug("is_valid_nhs_number: check digit mismatch")
        return False
    # Hooray!
    return True


def generate_random_nhs_number(official_test_range: bool = True) -> int:
    """
    Returns a random valid NHS number, as an ``int``.

    Args:
        official_test_range:
            Make it start with "999", the official NHS test range. See
            https://digital.nhs.uk/services/e-referral-service/document-library/synthetic-data-in-live-environments,
            saved at
            https://web.archive.org/web/20210116183039/https://digital.nhs.uk/services/e-referral-service/document-library/synthetic-data-in-live-environments.
    """  # noqa
    check_digit = 10  # NHS numbers with this check digit are all invalid
    while check_digit == 10:
        if official_test_range:
            # Make it start with 999.
            digits = [9, 9, 9]  # don't start with a zero
            digits.extend([random.randint(0, 9) for _ in range(6)])
        else:
            # Any NHS number. Could potentially be a real one.
            digits = [random.randint(1, 9)]  # don't start with a zero
            digits.extend([random.randint(0, 9) for _ in range(8)])
        # ... length now 9
        check_digit = nhs_check_digit(digits)
    # noinspection PyUnboundLocalVariable
    digits.append(check_digit)
    return int("".join([str(d) for d in digits]))


def test_nhs_rng(n: int = 100) -> None:
    """
    Tests the NHS random number generator.
    """
    for i in range(n):
        x = generate_random_nhs_number()
        assert is_valid_nhs_number(x), f"Invalid NHS number: {x}"


def generate_nhs_number_from_first_9_digits(first9digits: str) -> Optional[int]:
    """
    Returns a valid NHS number, as an ``int``, given the first 9 digits.
    The particular purpose is to make NHS numbers that *look* fake (rather
    than truly random NHS numbers which might accidentally be real).

    For example:

    .. code-block:: none

        123456789_ : no; checksum 10
        987654321_ : yes, valid if completed to 9876543210
        999999999_ : yes, valid if completed to 9999999999

    But see also :func:`generate_random_nhs_number` with its option to use the
    official NHS test range.
    """
    if len(first9digits) != 9:
        log.warning("Not 9 digits")
        return None
    try:
        first9int = int(first9digits)
    except (TypeError, ValueError):
        log.warning("Not an integer")
        return None  # not an int
    if len(str(first9int)) != len(first9digits):
        # e.g. leading zeros, or some such
        log.warning("Leading zeros?")
        return None
    check_digit = nhs_check_digit(first9digits)
    if check_digit == 10:  # NHS numbers with this check digit are all invalid
        log.warning("Can't have check digit of 10")
        return None
    return int(first9digits + str(check_digit))


# =============================================================================
# Get an NHS number out of text
# =============================================================================

WHITESPACE_REGEX = re.compile(r'\s')
NON_NUMERIC_REGEX = re.compile("[^0-9]")  # or "\D"


def nhs_number_from_text_or_none(s: str) -> Optional[int]:
    """
    Returns a validated NHS number (as an integer) from a string, or ``None``
    if it is not valid.

    It's a 10-digit number, so note that database 32-bit INT values are
    insufficient; use BIGINT. Python will handle large integers happily.

    NHS number rules:
    https://www.datadictionary.nhs.uk/version2/data_dictionary/data_field_notes/n/nhs_number_de.asp?shownav=0
    """  # noqa
    # None in, None out.
    funcname = "nhs_number_from_text_or_none: "
    if not s:
        log.debug(funcname + "incoming parameter was None")
        return None

    # (a) If it's not a 10-digit number, bye-bye.

    # Remove whitespace
    s = WHITESPACE_REGEX.sub("", s)  # replaces all instances
    # Contains non-numeric characters?
    if NON_NUMERIC_REGEX.search(s):
        log.debug(funcname + "contains non-numeric characters")
        return None
    # Not 10 digits long?
    if len(s) != 10:
        log.debug(funcname + "not 10 digits long")
        return None

    # (b) Validation
    n = int(s)
    if not is_valid_nhs_number(n):
        log.debug(funcname + "failed validation")
        return None

    # Happy!
    return n
